/*

  Copyright (c) 1998, 1999, Digital Creations, Fredericksburg, VA, USA.  
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are
  met:
  
    o Redistributions of source code must retain the above copyright
      notice, this list of conditions, and the disclaimer that follows.
  
    o Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions, and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.
  
    o Neither the name of Digital Creations nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.
  
  
  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.

  $Id: Buffer.c,v 1.2 1999/01/14 17:31:26 jim Exp $

  If you have questions regarding this software,
  contact:
 
    Digital Creations L.C.  
    info@digicool.com
 
    (540) 371-6909

*/
static char Buffer_module_documentation[] = 
"Provide 'buffer' objects that are essentially mutable strings.\n\n"
"\n$Id: Buffer.c,v 1.2 1999/01/14 17:31:26 jim Exp $"
;

#include "Python.h"

static void PyVar_Assign(PyObject **v, PyObject *e) { Py_XDECREF(*v); *v=e;}
#define ASSIGN(V,E) PyVar_Assign(&(V),(E))
#define UNLESS(E) if(!(E))
#define UNLESS_ASSIGN(V,E) ASSIGN(V,E); UNLESS(V)
#define OBJECT(O) ((PyObject*)(O))

#include "Buffer.h"

static void
Buffer_dealloc(Buffer *self)
{
  PyMem_DEL(self);
}

static PyObject *
Buffer_str(Buffer *self)
{
  return PyString_FromStringAndSize(self->val.c, self->ob_size);
}

static int
Buffer_length(Buffer *self)
{
  return self->ob_size/self->element_size;
}

static PyObject * 
Buffer_concat(Buffer *self, PyObject *bb)
{
  PyErr_SetString(PyExc_TypeError,
		  "Buffer objects do not support concatenation");
  return NULL;
}

static PyObject *
Buffer_repeat(Buffer *self, int n)
{
  PyErr_SetString(PyExc_TypeError,
		  "Buffer objects do not support repetition");
  return NULL;
}

static PyObject *
index_error(int i)
{
  PyObject *r;

  if((r=PyInt_FromLong(i)))
    {
      PyErr_SetObject(PyExc_IndexError, r);
      Py_DECREF(r);
    }
  return NULL;
}

static PyObject *
Buffer_item(Buffer *self, int i)
{
  int ii, s, es;

  es=self->element_size;
  s=self->ob_size/es;
  ii= i < 0 ? i+s : i;
  if(ii < 0 || ii >= s) return index_error(i);

  switch(self->element_type)
    {
    case 'c': 
      return PyString_FromStringAndSize(self->val.c+ii,1);
    case 's': 
      return PyString_FromStringAndSize(self->val.c+ii*es, es);
    case 'b':
      return PyInt_FromLong((long)(self->val.b[ii]));
    case 'B':			   
      return PyInt_FromLong((long)(self->val.B[ii]));
    case 'h':			   
      return PyInt_FromLong((long)(self->val.h[ii]));
    case 'H':			   		    
      return PyInt_FromLong((long)(self->val.H[ii]));
    case 'i':			   		    
      return PyInt_FromLong((long)(self->val.i[ii]));
    case 'I':			   		    
      return PyInt_FromLong((long)(self->val.I[ii]));
    case 'l':			   		    
      return PyInt_FromLong((long)(self->val.l[ii]));
    case 'L':			   		    
      return PyInt_FromLong((long)(self->val.L[ii]));
    case 'f':
      return PyFloat_FromDouble((double)(self->val.f[ii]));
    case 'd':
      return PyFloat_FromDouble(self->val.d[ii]);
    }
}

static PyObject *
Buffer_slice(Buffer *self, int ilow, int ihigh)
{
  int il, ih;

  if(self->element_type != 'c')
    {
      PyErr_SetString(PyExc_TypeError,
		      "Buffer slicing is only supported for character buffers."
		      );
      return NULL;
    }
    

  il= ilow < 0 ? ilow+self->ob_size : ilow;
  ih= ihigh < 0 ? ihigh+self->ob_size : ihigh;

  if(il < 0) il=0;
  else if(il > self->ob_size) il=self->ob_size;
  if(ih < 0) ih=0;
  else if(ih > self->ob_size) ih=self->ob_size;

  return PyString_FromStringAndSize(&(self->val.c[il]),ih-il);
}

static int 
buffer_ass(Buffer *self, int i, int l, PyObject *v)
{
  int vl;
  char *src;

  if(!v)
    {
      PyErr_SetString(PyExc_TypeError,
		      "deletion of buffer elements is not allowed");
      return -1;
    }

  if((vl=PyObject_Length(v)) < 0) return -1;
  if(vl != l)
    {
      PyErr_SetString(PyExc_TypeError,
		      "assignment would change the size of the buffer");
      return -1;
    }


  if(v->ob_type==BufferType) src=(((Buffer        *)v)->val.c);
  else if PyString_Check(v)  src=(((PyStringObject*)v)->ob_sval);
  else
    {
      PyErr_SetString(PyExc_TypeError,
		      "buffer assignment is only allowed from buffers\n"
		      "and strings.");
      return -1;
    }

  memcpy(self->val.c+i, src, l);
  return 0;
}

static int
Buffer_ass_float(Buffer *self, int i, PyObject *v)
{
  double d;
  char es;

  if(PyFloat_Check(v))
    d=PyFloat_AsDouble(v);
  else if((v=PyNumber_Float(v)))
    {
      d=PyFloat_AsDouble(v);
      Py_DECREF(v);
    }
  else return -1;
  
  es=self->element_size;
  if(self->element_type=='f') self->val.f[i]=d;
  else                        self->val.d[i]=d;
  
  return 0;
}

static int
Buffer_ass_item(Buffer *self, int i, PyObject *v)
{
  int ii, s, es;
  long d;

  es=self->element_size;
  s=self->ob_size/es;
  ii= i < 0 ? i+s : i;
  if(ii < 0 || ii >= s) return index_error(i), -1;

  if(!v)
    {
      PyErr_SetString(PyExc_TypeError,
		      "deletion of buffer elements is not allowed");
      return -1;
    }

  switch(self->element_type)
    {
    case 'c': 
      return buffer_ass(self, ii, 1, v);
    case 's':
      return buffer_ass(self, ii*es, es, v);
    case 'f':
    case 'd':
      return Buffer_ass_float(self, ii, v);
    }

  if(PyInt_Check(v))
    d=PyInt_AsLong(v);
  else if((v=PyNumber_Int(v)))
    {
      d=PyInt_AsLong(v);
      Py_DECREF(v);
    }
  else return -1;

  switch(self->element_type)
    {  
    case 'b':
      self->val.b[ii]=d;
      return 0;
    case 'B':
      self->val.B[ii]=d;
      return 0;
    case 'h':
      self->val.h[ii]=d;
      return 0;
    case 'H':
      self->val.H[ii]=d;
      return 0;
    case 'i':
      self->val.i[ii]=d;
      return 0;
    case 'I':
      self->val.I[ii]=d;
      return 0;
    case 'l':
      self->val.l[ii]=d;
      return 0;
    case 'L':
      self->val.L[ii]=d;
      return 0;
    }
}

static int
Buffer_ass_slice(Buffer *self, int ilow, int ihigh, PyObject *v)
{
  int il, ih;

  if(self->element_type != 'c')
    {
      PyErr_SetString(PyExc_TypeError,
		      "Buffer slicing is only supported for character buffers."
		      );
      return -1;
    }

  il= ilow < 0 ? ilow+self->ob_size : ilow;
  ih= ihigh < 0 ? ihigh+self->ob_size : ihigh;

  if(il < 0) il=0;
  else if(il > self->ob_size) il=self->ob_size;
  if(ih < 0) ih=0;
  else if(ih > self->ob_size) ih=self->ob_size;

  return buffer_ass(self, il, ih-il, v);
}

static PySequenceMethods Buffer_as_sequence = {
	(inquiry)Buffer_length,		/*sq_length*/
	(binaryfunc)Buffer_concat,		/*sq_concat*/
	(intargfunc)Buffer_repeat,		/*sq_repeat*/
	(intargfunc)Buffer_item,		/*sq_item*/
	(intintargfunc)Buffer_slice,		/*sq_slice*/
	(intobjargproc)Buffer_ass_item,	/*sq_ass_item*/
	(intintobjargproc)Buffer_ass_slice,	/*sq_ass_slice*/
};

static PyObject *
Buffer_GetAttr(Buffer *self, PyObject *name)
{
  char *n, t;

  if(n=PyString_AsString(name))
    {
      if(*n=='t' && strcmp(n,"typecode")==0)
	{
	  t=self->element_type;
	  return PyString_FromStringAndSize(&t,1);
	}
      if(*n=='i' && strcmp(n,"itemsize")==0)
	{
	  return PyInt_FromLong((long)(self->element_size));
	}      
    }
  PyErr_SetObject(PyExc_AttributeError, name);
  return NULL;
}

static PyTypeObject BufferType_ = {
  PyObject_HEAD_INIT(NULL) 0,
  "Buffer",			 /*tp_name*/
  sizeof(Buffer)-sizeof(double), /*tp_basicsize*/
  1,				/*tp_itemsize*/
  (destructor)Buffer_dealloc,	/*tp_dealloc*/
  (printfunc)0,	/*tp_print*/
  (getattrfunc)0,		/*obsolete tp_getattr*/
  (setattrfunc)0,		/*obsolete tp_setattr*/
  (cmpfunc)0,	/*tp_compare*/
  (reprfunc)0,		/*tp_repr*/
  0,		/*tp_as_number*/
  &Buffer_as_sequence,		/*tp_as_sequence*/
  0,		/*tp_as_mapping*/
  (hashfunc)0,		/*tp_hash*/
  (ternaryfunc)0,	/*tp_call*/
  (reprfunc)Buffer_str,		/*tp_str*/
  (getattrofunc)Buffer_GetAttr,	/*tp_getattro*/
  (setattrofunc)0,	/*tp_setattro*/
  
  0L,0L,
  "Mutable fixed-length data buffers"
};

static PyObject *
newBufferobject(PyObject *ignored, PyObject *args)
{
  int size;
  char etype='c';
  short esize=1;
  Buffer *self;

  UNLESS(PyArg_ParseTuple(args,"i|c",&size,&etype))
    {
      PyErr_Clear();
      UNLESS(PyArg_ParseTuple(args,"ih:Buffer",&size,&esize)) return NULL;
      etype='s';
    }
  
  switch(etype)
    {
    case 'c': 
    case 'b':
    case 'B':
      esize=sizeof(char); break;
    case 'h':
    case 'H':
      esize=sizeof(short); break;
    case 'i':
    case 'I':
      esize=sizeof(int); break;
    case 'l':
    case 'L':
      esize=sizeof(long); break;
    case 'f':
      esize=sizeof(float); break;
    case 'd':
      esize=sizeof(double); break;
    case 's':
      break;
    default:
      PyErr_SetString(PyExc_TypeError, "Invalid type code");
      return NULL;
    }

  UNLESS(self = PyObject_NEW_VAR(Buffer, BufferType,
				 size*esize)) return NULL;
  self->element_type=etype;
  self->element_size=esize;
  memset(self->val.c, 0, size*esize);
  return OBJECT(self);
}

static struct PyMethodDef Module_Level__methods[] = {
  {"Buffer", (PyCFunction)newBufferobject, METH_VARARGS,
   "Buffer(size) -- Create a new buffer"},
  {NULL, (PyCFunction)NULL, 0, NULL}		/* sentinel */
};

void
initBuffer()
{
  PyObject *m, *d;
  char *rev="$Revision: 1.2 $";

  m = Py_InitModule4("Buffer", Module_Level__methods,
		     Buffer_module_documentation,
		     (PyObject*)NULL,PYTHON_API_VERSION);

  d = PyModule_GetDict(m);

  BufferType=&BufferType_;
  BufferType_.ob_type=&PyType_Type;
  PyDict_SetItemString(d,"BufferType", OBJECT(BufferType));

  PyDict_SetItemString(d, "__version__",
		       PyString_FromStringAndSize(rev+11,strlen(rev+11)-2));
  
	
  if (PyErr_Occurred()) Py_FatalError("can't initialize module Buffer");
}

/*****************************************************************************
Revision Log:

  $Log: Buffer.c,v $
  Revision 1.2  1999/01/14 17:31:26  jim
  Changed copyright.

  Revision 1.1  1998/07/09 21:15:09  jim
  initial

  $Revision 1.2  1997/06/21 15:22:03  jim
  $Removed "object" suffix from generates C type names
  $
  $Revision 1.1  1997/02/24 23:25:42  jim
  $initial
  $

*****************************************************************************/
